import json
import traceback
import SimpleITK as sitk

from pathlib import Path
from loguru import logger


class Nifti:
    """A class used to represent a nifti file for mircat-v2 - including spacing, metadata, segmentations and others!"""

    def __init__(self, nifti_path: str | Path):
        nifti_path = Path(nifti_path).resolve()
        if not nifti_path.exists():
            message = f"Nifti file not found at {nifti_path}! Please check existence."
            logger.error(message)
            raise FileNotFoundError()
        if ".nii" not in nifti_path.suffixes:
            message = f"Input {nifti_path} not recognized as a nifti file. Please ensure file suffix is *.nii or *.nii.gz"
            logger.error(message)
            raise NotNiftiFileError()
        self.path = nifti_path
        self.parent = nifti_path.parent
        self.name = nifti_path.with_suffix("").stem
        self.seg_folder = self.parent / f"{self.name}_segs"
        self.segmentations = list(self.seg_folder.glob(f"{self.name}*.nii.gz"))
        try:
            self.img = sitk.ReadImage(nifti_path)
        except Exception as e:
            logger.error(
                f"Error loading {nifti_path} with SimpleITK :{e}\n{traceback.format_exc()}"
            )
            raise SimpleITKReadError()
        self.spacing = self.img.GetSpacing()
        self.shape = self.img.GetSize()
        if (nifti_path.parent / "metadata.json").exists():
            logger.debug(
                f"{nifti_path.parent / 'metadata.json'} found. Loading metadata."
            )
            with (nifti_path.parent / "metadata.json").open() as f:
                self.metadata: dict = json.load(f)
        else:
            logger.debug(
                f"No metadata.json file found next to {nifti_path}. Setting metadata to empty dictionary."
            )
            self.metadata: dict = {}


class SegNifti(Nifti):
    def __init__(self, nifti_path: str | Path):
        super().__init__(nifti_path)
        self.task_files: dict = {}

    def resample_and_save_for_segmentation(
        self, new_spacing, output_path, interpolator_type
    ) -> None:
        if len(new_spacing) == 2:
            # Set the new spacing z-length to be the same as the original
            new_spacing = [*new_spacing, self.spacing[-1]]
        # Resample image - always not a label
        resampled = resample_with_sitk(
            self.img,
            new_spacing=new_spacing,
            is_label=False,
            interpolator_type=interpolator_type,
        )
        sitk.WriteImage(resampled, output_path)
        logger.debug("Successfully resampled and wrote to {}", output_path)


class StatsNifti(Nifti):
    def __init__(self, nifti_path: str | Path, overwrite: bool = False):
        super().__init__(nifti_path)
        stats_file = self.seg_folder / f"{self.name}_stats.json"
        self.stats_file = stats_file
        if stats_file.exists() and not overwrite:
            logger.debug(f"Stats file {stats_file} found. Loading statistics.")
            with stats_file.open() as f:
                self.stats = json.load(f)
        else:
            if not overwrite:
                logger.debug(
                    f"No stats file found at {stats_file}. Setting stats to empty dictionary."
                )
            else:
                logger.debug(
                    f"Stats file {stats_file} exists but overwrite is set to True. Initializing empty stats."
                )
            self.stats = {}
        self.add_stats("nifti", str(self.path))
        if self.metadata != {}:
            logger.debug(f"Merging metadata with stats for {self.name}.")
            self.add_stats("series_uid", self.metadata.get("series_uid", "missing"))
            self.add_stats("metadata", dict(self.metadata))

    def set_task_to_id_map(self, task_to_id_map: dict) -> None:
        self.task_to_id_map = task_to_id_map
        return self

    def set_id_to_seg_map(self, id_to_seg_map: dict) -> None:
        """Set the mapping of task IDs to task names for this nifti."""
        self.id_to_seg_map = id_to_seg_map
        return self

    def set_vertebrae_midlines(self, vertebrae_stats: dict = None):
        """Set the vertebrae midlines for this nifti."""
        if vertebrae_stats is None:
            if "vertebrae" not in self.stats:
                logger.warning(
                    "No vertebrae stats found in nifti. Cannot set vertebrae midlines."
                )
                return self
            vertebrae_stats = self.stats["vertebrae"]
        vertebrae_midlines = {
            k: v.get("midline")
            for k, v in vertebrae_stats.items()
            if isinstance(v, dict)
        }
        self.vertebrae_midlines = vertebrae_midlines
        return self

    def add_stats(self, key: str, value: any) -> None:
        """Add a key-value pair to the stats dictionary."""
        if key in self.stats:
            logger.debug(
                f"Key {key} already exists in stats for {self.path}. Overwriting value."
            )
        self.stats[key] = value
        return self

    def preprocess_for_stats(
        self, new_spacing, image_resampler, label_resampler
    ) -> None:
        logger.debug(
            "Resampling image and segmentations for nifti {} to new spacing {}",
            self.name,
            new_spacing,
        )
        self.img = resample_with_sitk(
            self.img,
            new_spacing=new_spacing,
            is_label=False,
            interpolator_type=image_resampler,
            reorient="LAI",
        )
        self.spacing = self.img.GetSpacing()
        self.shape = self.img.GetSize()
        self.add_stats("stats_resolution", new_spacing)
        new_id_to_seg_map = {}
        for seg_id, seg_path in self.id_to_seg_map.items():
            logger.debug(
                "Resampling segmentation for seg_id {} at path {}", seg_id, seg_path
            )
            label = sitk.ReadImage(seg_path)
            resampled_label = resample_with_sitk(
                label,
                is_label=True,
                interpolator_type=label_resampler,
                reference_image=self.img,
            )
            assert resampled_label.GetSize() == self.img.GetSize(), (
                f"Resampled label size {resampled_label.GetSize()} does not match image size {self.img.GetSize()}"
            )
            new_id_to_seg_map[seg_id] = resampled_label
        self.set_id_to_seg_map(new_id_to_seg_map)
        return self

    def check_for_segmentations(self, task_list, task_map):
        task_to_id_map = {}
        id_to_segmentation_map = {}
        for task in task_list:
            allowed_segmentations = task_map.get(task, [])
            for seg_id in allowed_segmentations:
                seg_file = f"{self.name}_{seg_id}.nii.gz"
                if seg_file in [seg.name for seg in self.segmentations]:
                    task_to_id_map[task] = seg_id
                    id_to_segmentation_map[seg_id] = self.seg_folder / seg_file
        self.set_task_to_id_map(task_to_id_map)
        self.set_id_to_seg_map(id_to_segmentation_map)
        return self

    def save_json_stats(self):
        """Save the statistics to a JSON file."""
        save_stats = {k: v for k, v in self.stats.items() if "-db" not in k}
        with self.stats_file.open("w") as f:
            json.dump(save_stats, f, indent=4)


def resample_with_sitk(
    image,
    new_spacing=None,
    new_size=None,
    is_label=False,
    interpolator_type="gaussian",
    reference_image=None,
    reorient=None,
) -> sitk.Image:
    """Resample a SimpleITK image to a new spacing, size, or reference image.
    :param image: The SimpleITK image to resample.
    :param new_spacing: The new spacing to resample to. If None, the original
        spacing will be used to calculate the new size.
    :param new_size: The new size to resample to. If None, the original
        size will be used to calculate the new spacing.
    :param is_label: Whether the image is a label image. If True, the
        resampling will use label-specific interpolators.
    :param interpolator_type: The type of interpolator to use for resampling.
        Options are 'gaussian', 'linear', 'nearest' for label images, and
        'lanczos', 'bspline', 'gaussian', 'linear' for regular images.
    :param reference_image: An optional reference image to use for resampling.
        If provided, the new spacing and size will be taken from this image.
    :param reorient: The CT orientation to reorient the image to. If None, no reorientation is applied.
    :return: The resampled SimpleITK image."""
    # Get original properties
    original_spacing = image.GetSpacing()
    original_size = image.GetSize()

    resampler = sitk.ResampleImageFilter()
    if reference_image is not None:
        resampler.SetReferenceImage(reference_image)
    else:
        if new_spacing is not None and new_size is not None:
            raise ValueError("Only one of new_spacing and new_size may be specified")
        # Calculate new size based on spacing change
        if new_spacing is not None:
            new_size = [
                int(original_size[i] * original_spacing[i] / new_spacing[i])
                for i in range(3)
            ]
        elif new_size is not None:
            new_spacing = [
                original_spacing[i] * original_size[i] / new_size[i] for i in range(3)
            ]
        else:
            raise ValueError("Must specify either a new_size or new_spacing")
        #    Set up resampling parameters
        resampler.SetOutputSpacing(new_spacing)
        resampler.SetSize(new_size)
        resampler.SetOutputDirection(image.GetDirection())
        resampler.SetOutputOrigin(image.GetOrigin())
        resampler.SetTransform(sitk.Transform())
        resampler.SetDefaultPixelValue(0)

    if is_label:
        if interpolator_type == "gaussian":
            interpolator = sitk.sitkLabelGaussian
        elif interpolator_type == "linear":
            interpolator = sitk.sitkLabelLinear
        elif interpolator_type == "nearest":
            interpolator = sitk.sitkNearestNeighbor
        else:
            raise ValueError(
                f"label interpolator must be in [gaussian, linear, nearest], got {interpolator_type}"
            )
        resampler.SetInterpolator(interpolator)
    else:
        if interpolator_type == "lanczos":
            interpolator = sitk.sitkLanczosWindowedSinc
        elif interpolator_type == "bspline":
            interpolator = sitk.sitkBSpline
        elif interpolator_type == "gaussian":
            interpolator = sitk.sitkGaussian
        elif interpolator_type == "linear":
            interpolator = sitk.sitkLinear
        else:
            raise ValueError(
                f"image interpolator must be in [lanczos, gaussian, linear], got {interpolator_type}"
            )
        resampler.SetInterpolator(interpolator)
    resampled_image = resampler.Execute(image)
    if reorient is not None:
        current_orientation = (
            sitk.DICOMOrientImageFilter_GetOrientationFromDirectionCosines(
                resampled_image.GetDirection()
            )
        )
        if current_orientation != reorient:
            logger.debug(f"Reorienting image from {current_orientation} to {reorient}.")
            reorienter = sitk.DICOMOrientImageFilter()
            reorienter.SetDesiredCoordinateOrientation(reorient)
            resampled_image = reorienter.Execute(resampled_image)
    return resampled_image


class NotNiftiFileError(ValueError):
    pass


class SimpleITKReadError(ValueError):
    pass



import json
import sys
from math import floor
from multiprocessing import cpu_count
from pathlib import Path
from loguru import logger

__all__ = [
    "config_path",
    "dbase_config_file",
    "segmentation_config_file",
    "read_segmentation_config",
    "read_dbase_config",
    "logger_setup",
    "set_threads_per_process",
]

config_path = Path(__file__).parent.resolve()
dbase_config_file = config_path / "dbase.json"
segmentation_config_file = config_path / "segmentation_models.json"


def read_segmentation_config():
    if not segmentation_config_file.exists():
        logger.error(
            f"Segmentation config file {segmentation_config_file} does not exist. Add some segmentation models with mircat-v2 models add"
        )
        sys.exit(1)
    with segmentation_config_file.open() as f:
        return json.load(f)


def read_dbase_config():
    if not dbase_config_file.exists():
        logger.error(
            f"DBase config file {dbase_config_file} does not exist. Please check path or use mircat-v2 dbase create"
        )
        sys.exit(1)
    with dbase_config_file.open() as f:
        return json.load(f)


def logger_setup(verbose: bool, quiet: bool) -> None:
    """Set up logger for mircat-v2
    :param verbose: be verbose in the output by adding debug to stdout
    :param quiet: be quiet in the output by only showing successes and errors - no warnings or info.
    """
    logger.remove()
    # Regular
    stdout_fmt = "<green>{time: DD-MM-YYYY -> HH:mm:ss}</green> | <level>{level}</level> | <level>{message}</level>"
    stderr_fmt = "<red>{time: DD-MM-YYYY -> HH:mm:ss}</red> | <level>{level}</level> | <level>{message}</level>"
    if quiet:
        # Only show success messages and error messages
        logger.add(
            sys.stdout,
            format=stdout_fmt,
            level="SUCCESS",
            filter=lambda record: record["level"].no <= 25,
            enqueue=True,
        )
        logger.add(sys.stderr, format=stderr_fmt, level="ERROR", enqueue=True)
    elif verbose:
        # Include debugging output
        logger.add(
            sys.stdout,
            format=stdout_fmt,
            level="DEBUG",
            filter=lambda record: record["level"].no <= 25,
            enqueue=True,
        )
        logger.add(
            sys.stderr,
            format=stderr_fmt,
            level="WARNING",
            enqueue=True,
        )
    else:
        # Show everything above INFO
        logger.add(
            sys.stdout,
            format=stdout_fmt,
            level="INFO",
            filter=lambda record: record["level"].no <= 25,
            enqueue=True,
        )
        logger.add(
            sys.stderr,
            format=stderr_fmt,
            level="WARNING",
            enqueue=True,
        )


def set_threads_per_process(args):
    # Do some thread matching
    total_threads = cpu_count()
    required_threads = args.n_processes * args.threads_per_process
    logger.debug(
        f"Total threads on machine: {total_threads}. Requested threads for segmentation: {required_threads}"
    )
    if total_threads < required_threads:
        args.threads_per_process = floor(total_threads / args.n_processes)
        logger.warning(
            "Desired threads (n_processes * threads_per_process) > total threads on device ({}>{}). Limiting threads per worker to {}. If performance drops, consider reducing the number of workers and upping threads to match.",
            required_threads,
            total_threads,
            args.threads_per_process,
        )



import SimpleITK as sitk
from skimage.measure import label, regionprops


def _calculate_shape_stats(seg: sitk.Image) -> sitk.LabelShapeStatisticsImageFilter:
    """Calculate the shape stats for a segmentation
    Parameters
    ----------
    seg : sitk.Image
        The segmentation to calculate shape stats for
    Returns
    -------
    sitk.LabelShapeStatisticsImageFilter
        The executed shape stats
    """
    shape_stats = sitk.LabelShapeStatisticsImageFilter()
    shape_stats.SetGlobalDefaultCoordinateTolerance(1e-5)
    shape_stats.ComputeOrientedBoundingBoxOn()
    shape_stats.ComputePerimeterOn()
    shape_stats.Execute(seg)
    return shape_stats


def _calculate_intensity_stats(
    image: sitk.Image, seg: sitk.Image
) -> sitk.LabelIntensityStatisticsImageFilter:
    """Calculate intensity stats for a segmentation using the reference image
    Parameters
    ----------
    image : sitk.Image
        The reference image
    seg : sitk.Image
        The segmentation of the reference image
    Returns
    -------
    sitk.LabelIntensityStatisticsImageFilter
        The executed intensity stats
    """
    intensity_stats = sitk.LabelIntensityStatisticsImageFilter()
    intensity_stats.SetGlobalDefaultCoordinateTolerance(1e-5)
    intensity_stats.Execute(seg, image)
    return intensity_stats


def filter_segmentation(
    segmentation: sitk.Image,
    labels: str | list[str],
    label_map: dict[str, int],
    largest_connected_component: bool = False,
    remap: bool = False,
) -> sitk.Image:
    """Filter a segmentation to only include specified labels
    Parameters
    ----------
    segmentation : sitk.Image
        The segmentation to filter
    labels : str|list[str]
        The labels to keep in the segmentation
    label_map : dict[str, int]
        A mapping of structure names to label IDs
    remap : bool, optional
        If True, remap the labels indexed starting from 1, by default False
    Returns
    -------
    sitk.Image
        The filtered segmentation
    """
    filtered_segmentation = sitk.Image(segmentation.GetSize(), sitk.sitkUInt8)
    filtered_segmentation.CopyInformation(segmentation)
    if isinstance(labels, str):
        labels = [labels]
    for i, label in enumerate(labels, start=1):
        idx = label_map.get(label, None)
        if idx is not None:
            mask = sitk.BinaryThreshold(segmentation, idx, idx, 1, 0)
            if largest_connected_component:
                mask = get_largest_connected_component(mask)
            new_idx = i if remap else idx
            mask = mask * new_idx
            filtered_segmentation = sitk.Add(filtered_segmentation, mask)
    return filtered_segmentation


def get_largest_connected_component(binary_segmentation: sitk.Image) -> sitk.Image:
    """Get the largest connected component from a binary segmentation
    Parameters
    ----------
    binary_segmentation : sitk.Image
        The binary segmentation to get the largest connected component from
    Returns
    -------
    sitk.Image
        The largest connected component of the binary segmentation
    """
    cc_filter = sitk.ConnectedComponentImageFilter()
    labeled = cc_filter.Execute(binary_segmentation)
    relabel_filter = sitk.RelabelComponentImageFilter()
    relabeled = relabel_filter.Execute(labeled)
    largest_component = sitk.BinaryThreshold(relabeled, 1, 1, 1, 0)
    return largest_component


def get_regionprops(image) -> list:
    """Get region properties from a labeled image
    Parameters
    ----------
    image : np.ndarray
        The labeled image to get region properties from
    Returns
    -------
    list
        A list of region properties for each labeled region in the image
    """
    labeled_image = label(image)
    return regionprops(labeled_image)



import SimpleITK as sitk

from loguru import logger
from mircat_v2.stats.utilities import filter_segmentation
from mircat_v2.stats.vessel import Vessel


class Aorta(Vessel):
    """Class for aorta-specific statistics and operations for the custom segmentation model specifically."""

    anatomical_regions_vertebrae_map = {
        "thoracic": [f"T{i}" for i in range(3, 13)],
        "descending": [f"T{i}" for i in range(5, 13)],
        "upper_abdominal": ["T12", "L1", "L2"],
        "lower_abdominal": ["L2", "L3", "L4", "L5", "S1"],
    }

    def __init__(
        self,
        segmentation: sitk.Image,
        vertebrae_midlines: dict[str, int],
        label_map: dict,
    ):
        """
        Initialize the Aorta class with a nifti file and segmentation ID.
        :param nifti: StatsNifti object containing the nifti file and its metadata.
        :param seg_id: The segmentation ID to calculate statistics for.
        :param label_map: Dictionary mapping structure names to label IDs.
        """
        # We need to include the brachiocephalic trunk and left subclavian artery to separate anatomical regions
        labels = ["aorta", "brachiocephalic_trunk", "subclavian_artery_left"]
        self.segmentation = filter_segmentation(
            segmentation,
            labels=labels,
            label_map=label_map,
            remap=True,
            largest_connected_component=True,
        )
        self.anisotropy = list(self.segmentation.GetSpacing())
        self.array = sitk.GetArrayFromImage(self.segmentation)
        self.label_map = {idx: label for idx, label in enumerate(labels, start=1)}
        self.label = "aorta"
        self.label_idx = 1
        self.skeletonization_kwargs = {
            "teasar_params": {
                "scale": 1.0,
                "const": 40,
            },
            "object_ids": [1],  # Aorta is always label 1 in the segmentation
            "anisotropy": self.anisotropy,
            "dust_threshold": 1000,
            "progress": False,
            "fix_branching": True,
            "in_place": False,
            "fix_borders": True,
            "parallel": 1,
            "parallel_chunk_size": 100,
            "extra_targets_before": [],
            "extra_targets_after": [],
            "fill_holes": False,
            "fix_avocados": False,
            "voxel_graph": None,
        }
        self.vertebrae_midlines = vertebrae_midlines
        self._regions = None
        self.skeleton = None
        self.centerline = None
        self.cpr = None

    @property
    def regions(self):
        """
        Get the anatomical regions of the aorta.
        :return: Dictionary containing the anatomical regions and their properties.
        """
        if self._regions is None:
            self._determine_anatomical_regions()
        return self._regions

    def _determine_anatomical_regions(self):
        """
        Determine the anatomical regions of the aorta in the segmentation based on vertebrae midlines.
        """
        regions = {}
        for region, vertebrae in self.anatomical_regions_vertebrae_map.items():
            midlines = [
                self.vertebrae_midlines[vert]
                for vert in vertebrae
                if self.vertebrae_midlines.get(vert) is not None
            ]
            # Need at least two midlines to determine a region
            if len(midlines) > 1:
                regions[region] = {
                    "in_image": True,
                    "entire_region": len(midlines) == len(vertebrae),
                    "start": min(midlines),
                    "end": max(midlines),
                }
            else:
                regions[region] = {
                    "in_image": False,
                    "entire_region": False,
                    "start": None,
                    "end": None,
                }
        self._regions = regions
        return self


def calculate_aorta_stats(nifti, seg_id: str, label_map: dict) -> dict:
    """Calculate aorta specific statistics for a given nifti file.
    :param nifti: StatsNifti object containing the nifti file and its metadata.
    :param seg_id: The segmentation ID to calculate statistics for.
    :param label_map: Dictionary mapping structure names to label IDs.
    :return: Dictionary containing aorta specific statistics.
    """
    logger.debug(f"Calculating aorta stats for {nifti.name}.")
    if not hasattr(nifti, "vertebrae_midlines"):
        logger.warning(
            f"Vertebrae midlines not found in {nifti.name}. Skipping aorta stats calculation."
        )
        return {}
    segmentation = nifti.id_to_seg_map.get(seg_id)
    aorta = Aorta(
        segmentation,
        nifti.vertebrae_midlines,
        label_map,
    )



import sys
import json
import pprint

from math import floor
from multiprocessing import cpu_count
from pathlib import Path
from shutil import copytree
from loguru import logger

from mircat_v2.configs import (
    read_segmentation_config,
    read_dbase_config,
    segmentation_config_file,
    set_threads_per_process,
)
from mircat_v2.segmentation.segmentor import MircatSegmentor

lib_path = Path(__file__).parent.resolve()
models_path = lib_path / "models"


def add_segmentation_subparser(subparsers):
    """Add segmentation specific subcommands to mircat-v2
    :param subparsers: the subparser for CLI
    """
    # Add subcommands for segmentation itself
    seg_parser = subparsers.add_parser(
        "segment",
        description="""Segment nifti files using nnUNet - accelerated by GPU if available. \
            Input files will be resampled and stored in temporary files to accelerate nnUNet. \
            Make sure you have enough disk space for resampled images!""",
        help="Segment nifti files using nnUNet models.",
    )
    seg_parser.add_argument(
        "niftis",
        type=Path,
        help="Path to a nifti file or a text file containing a list of nifti files",
    )
    seg_parser.add_argument(
        "-tl",
        "--task-list",
        type=str,
        nargs="+",
        required=True,
        help="Space separated list of segmentation model tasks to perform. Identified by nnUNet dataset number. Use mircat-v2 models list to see all available tasks.",
    )
    seg_parser.add_argument(
        "-m",
        "--model-types",
        type=str,
        nargs="+",
        default=["3d"],
        choices=["2d", "3d", "3d_lowres", "3d_cascade"],
        help="Default = 3d for all tasks. The nnUNet model subtype to run for each task, space separated.",
    )
    seg_parser.add_argument(
        "-n",
        "--n-processes",
        type=int,
        default=1,
        help="Default = 1. Number of processes to use for pre/post processing the images.",
    )
    seg_parser.add_argument(
        "-t",
        "--threads-per-process",
        type=int,
        default=4,
        help="Default = 4. Maximum number of threads each process should use.",
    )
    seg_parser.add_argument(
        "-c",
        "--cache-size",
        type=int,
        default=10,
        help="The number of nifti files to work on at one time. Default = 10. This includes saving preprocessed files to disk, so be mindful of storage.",
    )
    seg_parser.add_argument(
        "-db",
        "--dbase-insert",
        action="store_true",
        help="Store model results (completed/failed for each nifti) in the mircat-v2 database. Must be setup!",
    )
    seg_parser.add_argument(
        "-ir",
        "--image-resampler",
        type=str,
        choices=["lanczos", "bspline", "gaussian", "linear"],
        default="bspline",
        help="Interpolator for resampling original images. Default = bspline (speed and quality balance)",
    )
    seg_parser.add_argument(
        "-lr",
        "--label-resampler",
        type=str,
        choices=["gaussian", "linear", "nearest"],
        default="gaussian",
        help="Interpolator for resampling segmentation images. Default = gaussian (slowest but best)",
    )
    # Model specific operations
    models_parser = subparsers.add_parser(
        "models",
        description="""List models available, copy new models to the correct mircat-v2 location, or update the config file for segmentation""",
        help="List, add or update models available to mircat-v2.",
    )
    models_subparser = models_parser.add_subparsers(dest="models_command")
    list_subparser = models_subparser.add_parser(
        "list",
        help="List tasks and their descriptions currently available to mircat-v2.",
    )
    list_subparser.add_argument(
        "-t",
        "--task",
        type=str,
        default="all",
        help="Specific task number to list configuration for. Default is to show all tasks.",
    )
    add_command = models_subparser.add_parser(
        "add",
        description="Add nnUNet model(s) to mircat-v2 options.",
        help="Add nnUNet model or folder of nnUNet models to mircat library",
    )
    add_command.add_argument(
        "folder",
        type=Path,
        help="Path to nnUNet model or folder containing multiple nnUNet models. \
            Should either be in the format DatasetXXX_* or be a folder containing folders in that format.",
    )
    add_command.add_argument(
        "--overwrite",
        action="store_true",
        help="If a model task already exists in the mircat models file, overwrite it.",
    )
    models_subparser.add_parser("update", help="Update the model configurations file.")


def segment_nifti_files(args):
    logger.debug(
        "Starting segmentation process with the following args:\n{}",
        pprint.pformat(args),
    )
    task_configs = read_segmentation_config()
    _validate_segmentation_args(task_configs, args)
    segmentor = MircatSegmentor(
        task_list=args.task_list,
        model_types=args.model_types,
        task_configs=task_configs,
        n_processes=args.n_processes,
        threads_per_process=args.threads_per_process,
        cache_size=args.cache_size,
        dbase_config=args.dbase_config,
        img_resampler=args.image_resampler,
        lbl_resampler=args.label_resampler,
    )
    segmentor.run(args.niftis)


def _get_available_models():
    task_folders = sorted(models_path.glob("Dataset*/"))
    logger.debug(
        """Found {} models in {}. They are:\n{}""",
        len(task_folders),
        str(models_path),
        pprint.pformat([str(x.name) for x in task_folders]),
    )
    if len(task_folders) == 0:
        raise FileNotFoundError(
            f"No folders found in {models_path}. Use mircat-v2 models copy to copy in some nnUNet models!"
        )
    return task_folders


def _validate_segmentation_args(task_configs: dict, args):
    """Internal function to validate the given arguments for nnUNet segmentation
    :param model_configs: the internal configuration file for mircat-v2 loaded as a dictionary
    :param args: the passed input parameters
    """
    # Make sure the input argument for the nifti file(s) exists
    args.niftis = args.niftis.resolve()
    if not args.niftis.exists():
        logger.error(
            f"The input nifti file/list of files {args.niftis} does not exist. Please double check your paths."
        )
        sys.exit(1)
    available_models = list(task_configs.keys())
    # Make sure all tasks are available
    if not all([task in available_models for task in args.task_list]):
        missing_tasks = [
            task for task in args.task_list if task not in available_models
        ]
        logger.error(
            "The following tasks are missing from mircat-v2 config file {}. Please use `mircat-v2 models add` to place them in the correct location.",
            missing_tasks,
        )
        sys.exit(1)
    # If only one model type was given, apply it for all tasks
    if len(args.model_types) == 1:
        model_type = args.model_types[0]
        logger.debug(
            "One model type passed to --model-types. Will apply for all given tasks."
        )
        args.model_types = [model_type for task in args.task_list]
    # Ensure that each task has a specified model type
    if len(args.model_types) != len(args.task_list):
        logger.error(
            f"Number of model type parameters must match the number of tasks. Currently have {len(args.model_types)} model types and {len(args.task_list)} tasks.\n\t\
            Please either specify one model type [3d, 2d, 3d_lowres, 3d_casade] for all tasks or have a specific model type for each task."
        )
        sys.exit(1)
    # Ensure that each task has weights for the specific model type
    for task, model in zip(args.task_list, args.model_types):
        model_weights = task_configs[task]["models"].get(model, dict())
        logger.debug(f"{task} configurations: {model_weights}")
        if not model_weights:
            logger.error(f"No {model} configuration found for task {task}.")
            sys.exit(1)
    # Check if the mircat-v2 dbase exists.
    if args.dbase_insert:
        dbase_config = read_dbase_config()
        dbase_path = Path(dbase_config["dbase_path"]).resolve()
        if not dbase_path.exists():
            logger.error(
                f"Database was set to {dbase_path}, but was not found. Please check path!"
            )
            sys.exit(1)
        logger.debug(f"mircat-v2 database found at {dbase_path}")
        args.dbase_config = dbase_config
    else:
        args.dbase_config = {}

    set_threads_per_process(args)


def update_models_config():
    # We call them tasks as this is the nnUNet style - Dataset###_descriptor
    task_folders = _get_available_models()
    logger.info(f"Updating {len(task_folders)} models configurations.")
    if not segmentation_config_file.parent.exists():
        logger.debug("Segmentation config file does not exist. Creating it.")
        config_data = {}
    else:
        logger.debug(
            f"Reading existing segmentation config file {segmentation_config_file}."
        )
        with segmentation_config_file.open() as f:
            config_data = json.load(f)
    # The keys here are what UNet calls each model type
    # The values are what we will use as flags to select a model
    unet_configs = {
        "2d": "2d",
        "3d_fullres": "3d",
        "3d_lowres": "3d_lowres",
        "3d_cascade_fullres": "3d_cascade",
    }
    for folder in task_folders:
        splits = folder.name.split("_")
        task = splits[0].replace("Dataset", "")
        description = "_".join(splits[1:])
        task_models = sorted(folder.glob("*/"))
        logger.debug(f"Task: {task}, Models: {task_models}")
        task_config = {"description": description, "models": {}}
        if len(task_models) == 0:
            logger.warning(f"No model options found in {folder}.")
            continue
        # Pull a reference model
        ref_model = task_models[0]
        if not (ref_model / "dataset.json").exists():
            logger.error(f"No dataset info found for {folder}. Skipping.")
            continue
        if not (ref_model / "plans.json").exists():
            logger.error(f"No model plans found for {folder}. Skipping.")
        with (ref_model / "dataset.json").open() as f:
            dataset = json.load(f)
        with (ref_model / "plans.json").open() as f:
            plans = json.load(f)
            configs = plans["configurations"]
        labels = dataset.get("labels")
        task_config["labels"] = labels
        # Get all of the nnUNet model options for the task that are available.
        for unet_config, option in unet_configs.items():
            for model in task_models:
                if unet_config in model.name:
                    model_specific_config = configs.get(unet_config)
                    task_config["models"][option] = {
                        "path": str(model),
                        "patch_size": model_specific_config.get("patch_size"),
                        "spacing": model_specific_config.get("spacing"),
                    }
        config_data[task] = task_config
    with segmentation_config_file.open("w") as f:
        json.dump(config_data, f, indent=4)
    logger.success(
        f"mircat-v2 model configs successfully updated. View them at {segmentation_config_file}."
    )


def add_models_to_mircat(args) -> None:
    folder: Path = args.folder
    overwrite: bool = args.overwrite
    if overwrite:
        confirmed = input(
            "Overwrite option was given to copy command. Are you sure you want to overwrite existing models? (y/n) "
        )
        confirmed = confirmed == "y"
        if not confirmed:
            logger.info("Aborting copy operation.")
            return

    folder = folder.resolve()
    if not folder.is_dir():
        logger.error(
            f"Given path {folder} is not a folder. Please make sure you are passing an nnUNet folder."
        )
        return
    if not any(folder.iterdir()):
        logger.error(
            f"Folder {folder} is empty. Please check that your path is correct."
        )
        return
    if folder.name.startswith("Dataset"):
        logger.info(f"Found 1 model to copy: {folder}")
        logger.info(f"Copying {folder} to {models_path}")
        destination = models_path / folder.name
        copytree(folder, destination, dirs_exist_ok=overwrite)
        logger.success(
            f"{folder} succesfully copied to mircat-v2 models. Updating config."
        )
    else:
        models = list(folder.glob("*/Dataset*"))
        n_models = len(models)
        if n_models == 0:
            logger.error(
                f"No models found in {folder}. Please make sure the input is either of format 'DatasetXXX_*' or folder/DatasetXXX_*"
            )
            return
        model_tasks = {
            model.name.split("_")[0].replace("Dataset", ""): model for model in models
        }
        logger.info(
            f"Found {n_models} models to copy in {folder} - Task numbers are {', '.join(model_tasks.keys())}"
        )
        for i, (task, model_path) in enumerate(model_tasks.items(), start=1):
            destination = models_path / model_path.name
            copytree(model_path, destination, dirs_exist_ok=overwrite)
            logger.success(
                f"{i}/{n_models} - Succesfully copied task {task} from {model_path} to mircat library."
            )
        logger.info("All models copied. Updating config.")
    update_models_config()


def list_mircat_models(args):
    with segmentation_config_file.open() as f:
        configs = json.load(f)
    if args.task == "all":
        pprint.pprint(configs, indent=2, sort_dicts=False)
        return
    else:
        subdict = configs.get(args.task)
        if subdict is None:
            logger.error(
                f"Task number {args.task} not found in model configs. Available tasks are\n\t{list(configs.keys())}",
            )
            return
        pprint.pprint(subdict, indent=2, sort_dicts=False)



import numpy as np

from dataclasses import dataclass, field
from loguru import logger
from scipy.interpolate import splprep, splev


@dataclass
class Centerline:
    """
    Class representing a centerline for a vessel.
    """

    label: int
    vessel: str
    coordinates: np.ndarray
    _differences: np.ndarray = field(init=False, repr=False, default=None)
    _segment_lengths: np.ndarray = field(init=False, repr=False, default=None)
    _cumulative_lengths: np.ndarray = field(init=False, repr=False, default=None)
    _length: float = field(init=False, repr=False, default=None)
    u: np.ndarray = field(init=True, repr=False, default=None)
    tck: np.ndarray = field(init=True, repr=False, default=None)

    def __call__(self):
        """
        Return the centerline coordinates.
        """
        return self.coordinates

    def __len__(self):
        """
        Return the number of points in the centerline.
        """
        return len(self.coordinates)
    
    def len(self) -> int:
        return len(self.coordinates)
    
    @property
    def differences(self) -> float:
        if self._differences is None:
            self._differences = np.diff(self.coordinates, axis=0)
        return self._differences

    @property
    def segment_lengths(self) -> np.ndarray:
        if self._segment_lengths is None:
            self._segment_lengths = np.sqrt(np.sum(self.differences**2, axis=1))
        return self._segment_lengths

    @property
    def cumulative_lengths(self) -> np.ndarray:
        if self._cumulative_lengths is None:
            self._cumulative_lengths = np.concatenate(
                ([0], np.cumsum(self.segment_lengths))
            )
        return self._cumulative_lengths

    @property
    def length(self) -> float:
        # This is the length in voxel space - not in physical space. Need to adjust for spacing if physical length is desired.
        if self._length is None:
            self._length = self.cumulative_lengths[-1]
        return self._length

    def _reset_properties(self):
        """
        Reset all cached properties to force recalculation.
        """
        self._differences = None
        self._segment_lengths = None
        self._cumulative_lengths = None
        self._length = None

    def resample_with_spline(
        self, target_points: int, smoothing: float = 0.5, degree: int = 3
    ):
        """
        Resample centerline using B-spline interpolation for smoother results.

        Args:
            target_points: Number of points in resampled centerline
            smoothing: Smoothing factor (0 = interpolation, >0 = approximation)
            degree: Spline degree (1=linear, 2=quadratic, 3=cubic)
        """
        if target_points <= 0:
            raise ValueError("Target number of points must be greater than zero.")

        if len(self.coordinates) < degree + 1:
            raise ValueError(
                f"Need at least {degree + 1} points for degree {degree} spline"
            )

        # Prepare coordinates for splprep (transpose for per-dimension arrays)
        centerline = self.coordinates
        coords_t = centerline.T
        # Get the u parameter
        dists = np.linalg.norm(np.diff(centerline, axis=0), axis=1)
        u = np.concatenate(([0], np.cumsum(dists)))
        u /= u[-1]
        # Fit B-spline to the centerline
        # u is the parameter array, automatically computed based on chord length
        s = smoothing * len(centerline)
        tck, u = splprep(coords_t, u=u, s=s, k=degree, nest=-1)
        # Generate new parameter values for resampling
        u_new = np.linspace(0, 1, target_points)
        # Evaluate spline at new parameter values
        resampled_coords = splev(u_new, tck)
        resampled_coords = np.array(resampled_coords).T.round().astype(int)
        # Update the centerline coordinates
        self.coordinates = resampled_coords
        self.u = u_new
        self.tck = tck
        self._reset_properties()
        return self

    def calculate_tangent_vectors(self):
        """
        Calculate tangent vectors, normals, and binormals for the centerline.
        Returns:
            tangents: Tangent vectors at each point
            normals: Normal vectors at each point
            binormals: Binormal vectors at each point
        """
        if self.tck is None:
            raise ValueError(
                "Centerline must be resampled with a spline before calculating tangent vectors."
            )
        # Evaluate the spline to get the tangent vectors
        tangents = np.array(splev(self.u, self.tck, der=1)).T
        tangents /= np.linalg.norm(tangents, axis=1, keepdims=True)
        self.tangents = tangents
        return self



import json
import SimpleITK as sitk
from pathlib import Path
from loguru import logger
from multiprocessing import Pool
from threadpoolctl import threadpool_limits
from mircat_v2.configs import read_dbase_config, read_segmentation_config
from mircat_v2.nifti import StatsNifti
from mircat_v2.stats.vol_int import calculate_volume_and_intensity_stats
from mircat_v2.stats.contrast import predict_contrast
from mircat_v2.stats.vertebrae import calculate_vertebrae_stats


class Analyzer:
    """Analyzer class for processing and analyzing tasks with specified parameters."""

    # Map for resolution settings
    resolution_map = {
        "normal": [1.0, 1.0, 1.0],
        "high": [0.75, 0.75, 0.75],
        "highest": [0.5, 0.5, 0.5],
    }
    # Segmentations needed for specific tasks in order of preference
    task_map = {
        "vol-int": ["999"],
        "contrast": ["999"],
        "vertebrae": ["999"],
        "aorta": ["999"],
        "tissues": ["485", "481"],
        # FIXME - need to add body segmentation id to tissues
    }

    def __init__(
        self,
        task_list: list[str],
        resolution: str,
        image_resampler: str,
        label_resampler: str,
        n_processes: int,
        threads_per_process: int,
        dbase_insert: bool = False,
        overwrite: bool = False,
    ):
        if task_list == ["all"]:
            task_list = list(self.task_map.keys())
        # This orders the tasks by the order they are defined in task_map
        self.task_list = [task for task in self.task_map if task in task_list]
        self.resolution = self.resolution_map.get(resolution)
        self.image_resampler = image_resampler
        self.label_resampler = label_resampler
        self.n_processes = n_processes
        self.threads_per_process = threads_per_process
        # Read database configuration if dbase_insert is True
        self.dbase_insert = dbase_insert
        if dbase_insert:
            self.dbase_config = read_dbase_config()
        else:
            self.dbase_config = {}
        # Read segmentation configuration
        seg_config = read_segmentation_config()
        # We only need the labels for each task
        self.segmentation_labels = {
            task: seg_config[task]["labels"] for task in seg_config
        }
        self.overwrite = overwrite

    def run(self, niftis: Path | str) -> None:
        """Run the analysis on the provided NiFTi files.
        :param niftis: Path to a nifti file with mircat-v2 segmentations or a text file with a list of multiple mircat-v2 nifti files.
        """
        self._get_nifti_list(niftis)
        if self.n_processes > 1:
            logger.info(f"Running analysis with {self.n_processes} processes.")
        else:
            logger.info("Running analysis with single process.")
        logger.info(
            f"Using resolution: {self.resolution} mm, image resampler: {self.image_resampler}, label resampler: {self.label_resampler}."
        )
        logger.info(f"Tasks to run: {self.task_list}.")
        logger.info(f"Starting analysis on {self.total_niftis} nifti files.")
        batch_data = []
        with Pool(processes=self.n_processes) as pool:
            for i, nifti in enumerate(
                pool.imap_unordered(self.analyze, self.niftis), start=1
            ):
                logger.success(
                    f"Stats: [{i}/{self.total_niftis}] - Successfully calculated statistics for {nifti.name}."
                )
                batch_data.append(nifti.stats)
                if i % 100 == 0 and self.dbase_insert:
                    logger.info("inserting batch of 100 nifti files into the database.")
                    # TODO Logic to insert batch_data into the database
                    batch_data.clear()
        if self.dbase_insert and batch_data:
            logger.info("inserting remaining nifti files into the database.")
            # TODO Logic to insert remaining batch_data into the database
            logger.info(batch_data)

    def _get_nifti_list(self, niftis):
        niftis = Path(niftis).resolve()
        if not niftis.exists():
            raise FileNotFoundError(f"Nifti file {niftis} does not exist.")
        if niftis.suffix in [".nii", ".gz"]:
            self.niftis = [niftis]
        else:
            with niftis.open() as f:
                self.niftis = [Path(line.strip()).resolve() for line in f.readlines()]
        self.total_niftis = len(self.niftis)
        if self.total_niftis == 0:
            raise ValueError("No valid nifti files found in the provided path or file.")
        logger.info(f"Found {self.total_niftis} nifti files to process from {niftis}.")

    def analyze(self, nifti_path: Path | str) -> StatsNifti:
        """Analyze a single nifti file.
        :param nifti_path: Path to the nifti file to analyze.
        """
        sitk.ProcessObject.SetGlobalDefaultNumberOfThreads(self.threads_per_process)
        with threadpool_limits(self.threads_per_process):
            nifti = StatsNifti(nifti_path, self.overwrite)
            # Preprocess the nifti file for statistics
            (
                nifti.check_for_segmentations(
                    self.task_list, self.task_map
                ).preprocess_for_stats(
                    self.resolution, self.image_resampler, self.label_resampler
                )
            )
            self.analyze_tasks(nifti)
        nifti.save_json_stats()
        return nifti

    def analyze_tasks(self, nifti: StatsNifti) -> None:
        for task in self.task_list:
            seg_id = nifti.task_to_id_map[task]
            seg_labels = self.segmentation_labels[seg_id]
            match task:
                case "vol-int":
                    vol_int_stats = calculate_volume_and_intensity_stats(
                        nifti, seg_id, seg_labels
                    )
                    nifti.add_stats("vol-int", vol_int_stats)
                case "contrast":
                    if not nifti.stats.get("vol-int", {}):
                        logger.warning(
                            "Organ intensities are needed for contrast prediction. Skipping contrast task."
                        )
                        continue
                    contrast_stats = predict_contrast(nifti.stats["vol-int"])
                    nifti.add_stats("contrast", contrast_stats)
                case "vertebrae":
                    vertebrae_stats = calculate_vertebrae_stats(
                        nifti, seg_id, seg_labels
                    )
                    nifti.add_stats("vertebrae", vertebrae_stats)
                    nifti.set_vertebrae_midlines(vertebrae_stats)
                # TODO: Implement other tasks as needed



import json
import traceback
from datetime import date
from pathlib import Path
from multiprocessing import Pool
from shutil import copyfile

import dicom2nifti
import dicom2nifti.settings as settings
from dicom2nifti.convert_dir import _remove_accents
import numpy as np
from pydicom import dcmread
from loguru import logger
from threadpoolctl import threadpool_limits

from ..dbase import insert_data_batch, dbase_config_file


class DicomConverter:
    """Class to convert DICOM files to NIfTI format using the dicom2nifti library.
    This class provides options for resampling, validation, and parallel processing.
    The output NIfTI files are saved in a specified directory with subdirectories based on the input DICOM file's metadata.
    """

    def __init__(
        self,
        axial_only: bool = False,
        no_mip: bool = False,
        resample: bool = True,
        resample_spline_interpolation_order: int = 3,
        resample_padding: int = -1024,
        validate_orthogonal: bool = False,
        n_processes: int = 1,
        threads_per_process: int = 4,
        min_slice_count: int = 30,
        db_batch_size: int = 100,
    ):
        """
        Initialize the DicomConverter with optional resampling settings.

        Args:
            axial_only (bool): Whether to only convert axial series. Default is False.
            no_mip (bool): Whether to skip converting MIP series. Default is False.
            resample (bool): Whether to resample the DICOM images to standard LAS orientation. Default is True.
            resample_spline_interpolation_order (int): The order of the spline interpolation for resampling. Default is 3.
            resample_padding (int): The padding value for resampling. Default is -1024 (Hounsfield units).
            validate_orthogonal (bool): Whether to validate the orthogonality of the DICOM images. Default is False.
                If True, gantry tilted images will not be converted.
            n_processes (int): The number of processes to use for conversion. Default is 1.
            threads_per_process (int): The maximum number of threads to use for conversion in each process. Default is 4.
            min_file_count (int): The minimum number of DICOM files in a folder required to perform conversion. Default is 30.
            db_batch_size (int): The number of records to insert into the database at once if it exists. Default is 100.
        """
        # Set the conversion options
        self.axial_only = axial_only
        self.no_mip = no_mip
        self.n_processes = n_processes
        self.threads_per_process = threads_per_process
        self.min_slice_count = min_slice_count
        self.db_batch_size = db_batch_size
        # Set the resampling options for the dicom2nifti library
        settings.resample = resample
        settings.resample_spline_interpolation_order = (
            resample_spline_interpolation_order
        )
        settings.resample_padding = resample_padding
        settings.validate_orthogonal = validate_orthogonal
        if dbase_config_file.exists():
            logger.info("Database configuration file found. Loading...")
            with dbase_config_file.open("r") as f:
                self.dbase_config = json.load(f)
        else:
            self.dbase_config = {}

    def convert(
        self, dicom_folders: list[str | Path], output_directory: str | Path
    ) -> None:
        """
        Convert DICOM files to NIfTI format.

        Args:
            dicom_folders (list[str|Path]): A list of DICOM folders to convert.
            output_directory (str|Path): Directory to save the converted NIfTI files.
        """
        # Validate the input dicom folders
        if not isinstance(dicom_folders, list):
            raise TypeError("dicom_folders must be a list of folder paths.")
        num_folders = len(dicom_folders)
        if num_folders == 0:
            raise ValueError("dicom_folders list is empty.")
        # Create the base output directory if it doesn't exist
        self.output_directory = Path(output_directory).resolve()
        if not self.output_directory.exists():
            logger.info(f"Creating output directory: {self.output_directory}")
            self.output_directory.mkdir(parents=True, exist_ok=True)
        # Run the conversion in parallel using multiprocessing
        if self.n_processes > num_folders:
            logger.info(
                f"Number of processes ({self.n_processes}) is greater than number of folders ({num_folders}). Setting number of processes to {num_folders}."
            )
            self.n_processes = num_folders
        logger.info(
            f"Running in {self.n_processes} process{'es' if self.n_processes > 1 else ''}."
        )
        completed = 1
        metadata_batch = []
        with Pool(self.n_processes) as pool:
            # We use imap_unordered because we don't care about the order of completion
            for result in pool.imap_unordered(self.convert_folder, dicom_folders):
                if result is not None:
                    dicom_folder, output_nifti, metadata = result
                    metadata["output_nifti"] = str(output_nifti)
                    metadata["conversion_date"] = date.today().strftime("%Y-%m-%d")
                    logger.success(
                        f"{completed}/{num_folders} - Converted {dicom_folder} to {output_nifti}",
                        extra={"folder": dicom_folder, "output": output_nifti},
                    )
                    if self.dbase_config and metadata:
                        metadata_batch.append(metadata)
                        if len(metadata_batch) >= self.db_batch_size:
                            insert_data_batch(
                                self.dbase_config,
                                "conversions",
                                metadata_batch,
                            )
                            metadata_batch = []
                completed += 1
        # Insert any remaining metadata into the database
        if self.dbase_config and metadata_batch:
            insert_data_batch(
                self.dbase_config,
                "conversions",
                metadata_batch,
            )

    def convert_folder(self, dicom_folder: str | Path) -> tuple[str, str] | None:
        """
        Wrapper of _convert_folder to handle exceptions and logging.
        Args:
            dicom_folder (str): Path to the DICOM folder to convert.
        Returns:
            tuple[str, str]: The input DICOM folder and output nifti file path.
        """
        try:
            return self._convert_folder(dicom_folder)
        except Exception as e:
            logger.error(
                f"Error converting folder {dicom_folder}: {e}\n{traceback.format_exc()}",
                extra={"folder": dicom_folder, "error": str(e)},
            )
            return None

    def _convert_folder(self, dicom_folder: str | Path) -> tuple[str, str]:
        """
        Convert a single DICOM folder to NIfTI format.

        Args:
            dicom_folder (str): Path to the DICOM folder to convert.
        Returns:
            tuple[str, str]: The input DICOM folder and output nifti file path.
        """
        with threadpool_limits(limits=self.threads_per_process):
            dicom_folder = Path(dicom_folder).resolve()
            if not dicom_folder.exists():
                raise FileNotFoundError(f"DICOM folder does not exist: {dicom_folder}")
            if not dicom_folder.is_dir():
                raise NotADirectoryError(f"Path is not a directory: {dicom_folder}")
            dicom_files = list(dicom_folder.glob("*.dcm"))
            if len(dicom_files) == 0:
                logger.warning(f"No DICOM files found in folder: {dicom_folder}")
                return
            if len(dicom_files) < self.min_slice_count:
                logger.warning(
                    f"Skipping folder {dicom_folder} with less than {self.min_slice_count} DICOM files."
                )
                return
            # Get the metadata from the DICOM files
            reference_dicom = dicom_files[0]
            metadata = self._get_metadata(reference_dicom)
            if metadata.get("series_name") == "nan":
                logger.warning(
                    f"Skipping folder {dicom_folder} due to series_name == 'nan'."
                )
                return
            if "ai-rad_companion" in metadata.get("series_name"):
                logger.warning(
                    f"Skipping AI-Rad companion report series {metadata.get('original_series_name')} in folder {dicom_folder}."
                )
                return
            if self.axial_only and metadata.get("ct_direction") != "AX":
                logger.warning(
                    f"Skipping non-axial series {metadata.get('original_series_name')} in folder {dicom_folder}."
                )
                return
            if self.no_mip:
                # Do a check for MIP series
                if (
                    "mip" in metadata.get("series_name").lower()
                    or metadata.get("slice_thickness_mm", 5) > 5
                ):
                    logger.warning(
                        f"Skipping likely MIP series {metadata.get('original_series_name')} in folder {dicom_folder}."
                    )
                    return
            output_nifti = (
                f"{metadata['series_number']}_{metadata['series_name']}.nii.gz"
            )
            custom_output_directory = (
                self.output_directory
                / metadata.get("mrn")[:2]  # First 2 numbers of MRN
                / metadata.get("mrn")[2:4]  # Next 2 numbers of MRN
                / metadata.get("mrn")
                / metadata.get("accession")
                / metadata.get("series_name")
            )
            if not custom_output_directory.exists():
                custom_output_directory.mkdir(parents=True, exist_ok=True)
            # Convert the DICOM files to NIfTI format
            dicom2nifti.convert_directory(dicom_folder, custom_output_directory)
            # Save the metadata to a JSON file
            metadata_file = custom_output_directory / "metadata.json"
            with metadata_file.open("w") as f:
                json.dump(metadata, f, indent=4)
            # Copy the reference DICOM file to the output directory
            dicom_file = custom_output_directory / reference_dicom.name
            copyfile(reference_dicom, dicom_file)
            return dicom_folder, custom_output_directory / output_nifti, metadata

    def _get_metadata(self, reference_dicom: Path) -> dict:
        """Get the metadata from a DICOM file that is not saved in the NIfTI header.
        Args:
            reference_dicom (Path): Path to the reference DICOM file.
        Returns:
            dict: A dictionary containing the metadata.
        """
        orientations = {
            "COR": [1, 0, 0, 0, 0, -1],
            "SAG": [0, 1, 0, 0, 0, -1],
            "AX": [1, 0, 0, 0, 1, 0],
        }
        meta_attributes = {
            "PatientID": "mrn",
            "AccessionNumber": "accession",
            "SeriesDescription": "series_name",
            "SeriesNumber": "series_number",
            "StudyDescription": "study_description",
            "ImageOrientationPatient": "ct_direction",
            "ImageType": "image_type",
            "PatientSex": "sex",
            "PatientAge": "age",
            "PatientSize": "height_m",
            "PatientWeight": "weight_kg",
            "PregnancyStatus": "pregnancy_status",
            "PatientBirthDate": "birth_date",
            "AcquisitionDate": "scan_date",
            "PixelSpacing": None,
            "SliceThickness": "slice_thickness_mm",
            "Manufacturer": "manufacturer",
            "ManufacturerModelName": "model",
            "KVP": "kvp",
            "SequenceName": "sequence_name",
            "ProtocolName": "protocol_name",
            "ContrastBolusAgent": "contrast_bolus_agent",
            "ContrastBolusRoute": "contrast_bolus_route",
            "ContrastBolusVolume": "contrast_bolus_volume",
            "StudyInstanceUID": "study_uid",
            "SeriesInstanceUID": "series_uid",
            "Modality": "modality",
        }
        metadata = {}
        try:
            ref = dcmread(reference_dicom)
        except FileNotFoundError:
            logger.error(f"Reference DICOM file not found: {reference_dicom}")
            return metadata
        except Exception as e:
            logger.error(f"Error reading DICOM file {reference_dicom}: {e}")
            return metadata
        # Need these 3 fields to create the output directory
        if not hasattr(ref, "PatientID"):
            raise ValueError(
                f"Reference DICOM file does not have PatientID: {reference_dicom.parent}"
            )
        if not hasattr(ref, "AccessionNumber"):
            raise ValueError(
                f"Reference DICOM file does not have AccessionNumber: {reference_dicom.parent}"
            )
        if not hasattr(ref, "SeriesDescription"):
            raise ValueError(
                f"Reference DICOM file does not have SeriesDescription: {reference_dicom.parent}"
            )
        # Go through the metadata attributes and get the values
        for tag, column in meta_attributes.items():
            if not hasattr(ref, tag):
                # Not every DICOM file has all the tags, so we skip them
                continue
            value = getattr(ref, tag)
            match tag:
                case "PatientID":
                    value = str(value).zfill(0)
                case "SeriesDescription":
                    # Remove accents from the SeriesDescription
                    metadata["original_series_name"] = value
                    value = _remove_accents(value)
                case "SeriesNumber":
                    value = int(value)
                case "ImageOrientationPatient":
                    value = [round(float(x)) for x in value]
                    for orientation, array in orientations.items():
                        if np.allclose(value, array):
                            value = orientation
                            break
                    if isinstance(value, list):
                        value = "AX"
                    if isinstance(value, str) and value not in orientations:
                        value = "AX"
                case "ImageType":
                    value = "_".join(value)
                case tag if tag in ["PatientBirthDate", "AcquisitionDate"]:
                    value = date.fromisoformat(value).strftime("%Y-%m-%d")
                case "PatientAge":
                    value = int(value[:-1])
                case "PixelSpacing":
                    length, width = value
                    metadata["pixel_length_mm"] = float(length)
                    metadata["pixel_width_mm"] = float(width)
                    continue
                case "PregnancyStatus":
                    value = int(value)
                    if value == 4:
                        value = None
                    elif value == 1:
                        value = 0
                    elif value == 2:
                        value = 1
                    elif value == 3:
                        value = 2
                case tag if tag in [
                    "PatientSize",
                    "PatientWeight",
                    "KVP",
                    "SliceThickness",
                ]:
                    value = float(value)
                case tag if tag in ["SequenceName", "ProtocolName"]:
                    value = str(value).lower()
                case _:
                    pass
            if value == "":
                value = None
            metadata[column] = value
        if not metadata:
            raise ValueError(
                f"Reference DICOM file does not have any metadata: {reference_dicom.parent}"
            )
        if metadata.get("slice_thickness_mm") is None:
            metadata["slice_thickness_mm"] = 5
        metadata["dicom_folder"] = str(reference_dicom.parent)
        return metadata



import json
import sqlite3
from pathlib import Path

from loguru import logger
from mircat_v2.configs import dbase_config_file

MIRCAT_DB_TABLES = {
    "conversions": {
        "series_uid": "TEXT PRIMARY KEY",
        "study_uid": "TEXT",
        "modality": "TEXT",
        "mrn": "TEXT",
        "accession": "TEXT",
        "series_name": "TEXT",
        "series_number": "INTEGER",
        "scan_date": "TEXT",
        "original_series_name": "TEXT",
        "study_description": "TEXT",
        "ct_direction": "TEXT",
        "image_type": "TEXT",
        "sex": "TEXT",
        "age": "INTEGER",
        "birth_date": "TEXT",
        "height_m": "REAL",
        "weight_kg": "REAL",
        "pregnancy_status": "INTEGER",
        "pixel_length_mm": "REAL",
        "pixel_width_mm": "REAL",
        "slice_thickness_mm": "REAL",
        "manufacturer": "TEXT",
        "model": "TEXT",
        "kvp": "REAL",
        "sequence_name": "TEXT",
        "protocol_name": "TEXT",
        "contrast_bolus_agent": "TEXT",
        "contrast_bolus_route": "TEXT",
        "contrast_bolus_volume": "REAL",
        "dicom_folder": "TEXT",
        "nifti": "TEXT",
        "conversion_date": "TEXT",
    },
    "segmentations": {
        "nifti": "TEXT",
        "series_uid": "TEXT",
        "task": "INTEGER",
        "seg_file": "TEXT",
        "seg_date": "TEXT",
        "status": "TEXT",
        "failed_error": "TEXT",
        "PRIMARY KEY": "(nifti, task, seg_date)",
    },
}


def add_dbase_subparser(subparsers):
    # Add subcommands
    dbase_parser = subparsers.add_parser("dbase", help="Database management commands")
    dbase_subparsers = dbase_parser.add_subparsers(dest="dbase_command")

    # Add subcommand for creating a new database
    create_parser = dbase_subparsers.add_parser("create", help="Create a new database")
    create_parser.add_argument(
        "dbase_path",
        type=Path,
        help="Path to the database file",
    )
    create_parser.add_argument(
        "--overwrite",
        action="store_true",
        help="Overwrite the database if it already exists",
    )
    create_parser.add_argument(
        "--no-setup",
        action="store_true",
        help="Do not set up the database schema",
    )
    create_parser.add_argument(
        "--update",
        action="store_true",
        help="Update the database schema if it already exists",
    )

    # add subcommand for querying the database
    query_parser = dbase_subparsers.add_parser("query", help="Query the database")
    query_parser.add_argument(
        "query",
        type=str,
        help="SQL query to execute",
    )


def create_database(dbase_path, update=False, overwrite=False, no_setup=False):
    """
    Create a new SQLite database at the specified path.

    Args:
        dbase_path (Path): Path to the database file.
        update (bool): If True, update the database schema if it already exists.
        overwrite (bool): If True, overwrite the database if it already exists.
        no_setup (bool): If True, do not set up the database schema.
    """
    if dbase_path.exists() and not update and not overwrite:
        raise FileExistsError(
            f"Database file {dbase_path} already exists. Use --update to update the schema or --overwrite to completely overwrite."
        )
    elif dbase_path.exists() and update:
        logger.info(f"Updating existing database at {dbase_path}")
        conn = sqlite3.connect(dbase_path)
        cursor = conn.cursor()
        create_tables_from_schema(cursor, MIRCAT_DB_TABLES)
        conn.commit()
        conn.close()
        save_config(dbase_path)
        logger.success(f"Database updated at {dbase_path}")
        return
    elif dbase_path.exists() and overwrite:
        response = input(
            "Are you sure you want to overwrite the existing database? (y/n): "
        )
        if response.lower() != "y":
            logger.info(f"Left existing database at {dbase_path} intact.")
            return
        logger.info(f"Overwriting existing database at {dbase_path}")
        dbase_path.unlink()
    # Create the database
    conn = sqlite3.connect(dbase_path)
    logger.success(f"Database created at {dbase_path}")
    cursor = conn.cursor()

    if not no_setup:
        create_tables_from_schema(cursor, MIRCAT_DB_TABLES)

    conn.commit()
    conn.close()

    save_config(dbase_path)


def save_config(dbase_path):
    config = {"dbase_path": str(dbase_path.resolve()), "tables": MIRCAT_DB_TABLES}
    with dbase_config_file.open("w") as f:
        json.dump(config, f, indent=4)
    logger.success(f"Database configuration saved to {dbase_config_file}")


def create_tables_from_schema(cursor, schema_dict):
    """Create tables from a schema dictionary."""
    for table_name, columns in schema_dict.items():
        column_defs = [
            f"{col_name} {col_type}" for col_name, col_type in columns.items()
        ]
        create_table_sql = f"""
            CREATE TABLE IF NOT EXISTS {table_name} (
                {", ".join(column_defs)}
            )
        """
        cursor.execute(create_table_sql)
        logger.info(f"Created table: {table_name}")


def insert_data_batch(dbase_config: dict, table: str, data_records: list[dict]):
    """Insert a batch of data into the database for a specific table."""
    if not data_records:
        return
    dbase_path = Path(dbase_config["dbase_path"])
    conn = sqlite3.connect(dbase_path)
    cursor = conn.cursor()
    try:
        # Get the columns that exist in the table
        table_schema = dbase_config["tables"].get(table)
        if not table_schema:
            raise ValueError(
                f"Table {table} does not exist in the standard MirCAT database schema."
            )
        # Check if the table exists
        cursor.execute(
            f"SELECT name FROM sqlite_master WHERE type='table' AND name='{table}'"
        )
        if not cursor.fetchone():
            logger.info(f"Table {table} does not exist. Creating it.")
            create_tables_from_schema(cursor, {table: table_schema})
            conn.commit()
            logger.info(f"Table {table} created successfully.")
        table_columns = list(table_schema.keys())

        # Process each record in the batch
        insert_data = []
        for metadata in data_records:
            # Filter metadata to only include columns that exist in the table and
            # put in correct order. We use get to insert null values for missing columns
            insert_data.append(
                {
                    col: metadata.get(col)
                    for col in table_columns
                    if col != "PRIMARY KEY"
                }
            )
        logger.debug("Insert data: {}", insert_data)

        if insert_data:
            # Use the first record to determine column order
            columns = list(insert_data[0].keys())
            placeholders = ", ".join(["?" for _ in columns])
            insert_sql = (
                f"REPLACE INTO {table} ({', '.join(columns)}) VALUES ({placeholders})"
            )

            # Prepare values for batch insert
            values_list = []
            for metadata in insert_data:
                values = [metadata.get(col) for col in columns]
                values_list.append(values)

            # Execute batch insert
            cursor.executemany(insert_sql, values_list)
            conn.commit()
            logger.success(f"Inserted {len(values_list)} records into {table} table")
    except Exception as e:
        conn.rollback()
        logger.error(f"Error inserting batch: {e}")
        raise
    finally:
        conn.close()
